import random
import time

import numpy as np
from pymoo.core.population import Population
from pymoo.operators.survival.rank_and_crowding import RankAndCrowding
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from sklearn.naive_bayes import GaussianNB

from pydmoo.algorithms.base.dmoo.dmoeadde import DMOEADDE


def euclidean_distance(a, b):
    a = np.array(a)
    b = np.array(b)
    return np.sqrt(np.sum((a - b) ** 2))


class MOEADDEKGB(DMOEADDE):
    """Knowledge Guided Bayesian (KGB).

    References
    ----------
    Ye, Y., Li, L., Lin, Q., Wong, K.-C., Li, J., and Ming, Z. (2022).
    Knowledge guided bayesian classification for dynamic multi-objective optimization.
    Knowledge-Based Systems, 250, 109173.
    https://doi.org/10.1016/j.knosys.2022.109173
    """
    def __init__(
        self,
        c_size=13,
        perturb_dev=0.1,
        ps={},
        save_ps=False,
        **kwargs,
    ):

        super().__init__(**kwargs)

        self.C_SIZE = c_size
        self.PERTURB_DEV = perturb_dev

        self.ps = ps
        self.save_ps = save_ps

        self.nr_rand_solutions = 50 * self.pop_size
        self.t = 0

    def _next_static_dynamic(self):
        # for dynamic environment
        pop = self.pop

        if self.state is None:

            # archive the current POS
            self.add_to_ps()

            change_detected = self._detect_change_sample_part_population()

            if change_detected:

                start_time = time.time()

                pop = self._response_mechanism()

                # reevaluate because we know there was a change
                self.evaluator.eval(self.problem, self.pop, skip_already_evaluated=False, algorithm=self)

                # reevaluate because we know there was a change
                self.evaluator.eval(self.problem, pop)

                if len(pop) > self.pop_size:
                    # do a survival to recreate rank and crowding of all individuals
                    # Modified by DynOpt on Dec 21, 2025
                    # n_survive=len(pop) -> n_survive=self.pop_size
                    pop = RankAndCrowding().do(self.problem, pop, n_survive=self.pop_size, random_state=self.random_state)  # len(pop)

                self.data["response_duration"] = time.time() - start_time

        return pop

    def _response_mechanism(self):
        """Response mechanism."""
        # increase t counter for unique key of PS
        self.t += 1

        # conduct knowledge reconstruction examination
        pop_useful, pop_useless, c = self.knowledge_reconstruction_examination()

        # Train a naive bayesian classifier
        model = self.naive_bayesian_classifier(pop_useful, pop_useless)

        # generate a lot of random solutions with the dimensions of problem decision space
        X_test = self.random_strategy(self.nr_rand_solutions)

        # introduce noise to vary previously useful solutions
        noise = self.random_state.normal(0, self.PERTURB_DEV, self.problem.n_var)
        noisy_useful_history = np.asarray(pop_useful) + noise

        # check whether solutions are within bounds
        noisy_useful_history = self.check_boundaries(noisy_useful_history)

        # add noisy useful history to randomly generated solutions
        X_test = np.vstack((X_test, noisy_useful_history))

        # predict whether random solutions are useful or useless
        Y_test = model.predict(X_test)

        # create list of useful predicted solutions
        predicted_pop = self.predicted_population(X_test, Y_test)

        # ------ POPULATION GENERATION --------
        # take a random sample from predicted pop and known useful pop
        if len(predicted_pop) >= self.pop_size - self.C_SIZE:
            init_pop = []
            predicted_pop = random.sample(
                predicted_pop, self.pop_size - self.C_SIZE
            )

            # add sampled solutions to init_pop
            for solution in predicted_pop:
                init_pop.append(solution)

            # add cluster centroids to init_pop
            for solution in c:
                init_pop.append(np.asarray(solution))

        else:

            # if not enough predicted solutions are available, add all predicted solutions to init_pop
            init_pop = []

            for solution in predicted_pop:
                init_pop.append(solution)

            # add cluster centroids to init_pop
            for solution in c:
                init_pop.append(np.asarray(solution))

        # if there are still not enough solutions in init_pop randomly sample previously useful solutions directly without noise to init_pop
        if len(init_pop) < self.pop_size:

            # fill up init_pop with randomly sampled solutions from pop_useful
            if len(pop_useful) >= self.pop_size - len(init_pop):

                init_pop = np.vstack(
                    (
                        init_pop,
                        random.sample(pop_useful, self.pop_size - len(init_pop)),
                    )
                )
            else:
                # if not enough solutions are available, add all previously known useful solutions without noise to init_pop
                for solution in pop_useful:
                    init_pop.append(solution)

        # if there are still not enough solutions in init_pop generate random solutions with the dimensions of problem decision space
        if len(init_pop) < self.pop_size:

            # fill up with random solutions
            init_pop = np.vstack(
                (init_pop, self.random_strategy(self.pop_size - len(init_pop)))
            )

        # recreate the current population without being evaluated
        pop = Population.new(X=init_pop)

        # reevaluate because we know there was a change
        self.evaluator.eval(self.problem, pop)

        # do a survival to recreate rank and crowding of all individuals
        pop = RankAndCrowding().do(self.problem, pop, n_survive=len(pop), random_state=self.random_state)  # self.pop_size

        return pop

    def knowledge_reconstruction_examination(self):
        """
        Perform the knowledge reconstruction examination.
        :return: Tuple containing the useful population, useless population, and cluster centroids
        """
        clusters = self.ps  # set historical PS set as clusters
        Nc = self.C_SIZE  # set final nr of clusters
        size = len(self.ps)  # set size iteration to length of cluster
        run_counter = 0  # counter variable to give unique key

        # while there are still clusters to be condensed
        while size > Nc:

            counter = 0
            min_distance = None
            min_distance_index = []

            # get clusters that are closest to each other by calculating the euclidean distance
            for keys_i in clusters.keys():
                for keys_j in clusters.keys():
                    if (
                        clusters[keys_i]["solutions"]
                        is not clusters[keys_j]["solutions"]
                    ):

                        dst = euclidean_distance(
                            clusters[keys_i]["centroid"],
                            clusters[keys_j]["centroid"],
                        )

                        if min_distance == None:
                            min_distance = dst
                            min_distance_index = [keys_i, keys_j]
                        elif dst < min_distance:
                            min_distance = dst

                            min_distance_index = [keys_i, keys_j]

                        counter += 1

            # merge closest clusters
            for solution in clusters[min_distance_index[1]]["solutions"]:
                clusters[min_distance_index[0]]["solutions"].append(solution)

            # calculate new centroid for merged cluster
            clusters[min_distance_index[0]][
                "centroid"
            ] = self.calculate_cluster_centroid(
                clusters[min_distance_index[0]]["solutions"]
            )

            # remove cluster that was merged
            del clusters[min_distance_index[1]]

            size -= 1
            run_counter += 1

        c = []  # list of centroids
        pop_useful = []
        pop_useless = []

        # get centroids of clusters
        for key in clusters.keys():
            c.append(clusters[key]["centroid"])

        # create pymoo population objected to evaluate centroid solutions
        centroid_pop = Population.new("X", c)

        # evaluate centroids
        self.evaluator.eval(self.problem, centroid_pop)

        # do non-dominated sorting on centroid solutions
        ranking = NonDominatedSorting().do(centroid_pop.get("F"), return_rank=True)[-1]

        # add the individuals from the clusters with the best objective values to the useful population the rest is useless :(

        for idx, rank in enumerate(ranking):
            if rank == 0:
                for key in clusters.keys():
                    if centroid_pop[idx].X == clusters[key]["centroid"]:
                        for cluster_individual in clusters[key]["solutions"]:
                            pop_useful.append(cluster_individual)
            else:
                for key in clusters.keys():
                    if centroid_pop[idx].X == clusters[key]["centroid"]:
                        for cluster_individual in clusters[key]["solutions"]:
                            pop_useless.append(cluster_individual)

        # return useful and useless population and the centroid solutions
        return pop_useful, pop_useless, c

    def naive_bayesian_classifier(self, pop_useful, pop_useless):
        """
        Train a naive Bayesian classifier using the useful and useless populations.
        :param pop_useful: Useful population
        :param pop_useless: Useless population
        :return: Trained GaussianNB classifier
        """
        labeled_useful_solutions = []
        labeled_useless_solutions = []

        # add labels to solutions
        for individual in pop_useful:
            labeled_useful_solutions.append((individual, +1))

        for individual in pop_useless:
            labeled_useless_solutions.append((individual, -1))

        x_train = []
        y_train = []

        for i in range(len(labeled_useful_solutions)):
            x_train.append(labeled_useful_solutions[i][0])
            y_train.append(labeled_useful_solutions[i][1])

        for i in range(len(labeled_useless_solutions)):
            x_train.append(labeled_useless_solutions[i][0])
            y_train.append(labeled_useless_solutions[i][1])

        x_train = np.asarray(x_train)
        y_train = np.asarray(y_train)

        # fit the naive bayesian classifier with the training data
        model = GaussianNB()
        model.fit(x_train, y_train)

        return model

    def add_to_ps(self):
        """
        Add the current Pareto optimal set (POS) to the Pareto set (PS) with individual keys.
        """

        PS_counter = 0

        for individual in self.opt:

            if isinstance(individual.X, list):
                individual.X = np.asarray(individual.X)

            centroid = self.calculate_cluster_centroid(individual.X)

            self.ps[str(PS_counter) + "-" + str(self.t)] = {
                "solutions": [individual.X.tolist()],
                "centroid": centroid,
            }

            PS_counter += 1

    def predicted_population(self, X_test, Y_test):
        """
        Create a predicted population from the test set with positive labels.
        :param X_test: Test set of features
        :param Y_test: Test set of labels
        :return: Predicted population
        """
        predicted_pop = []
        for i in range(len(Y_test)):
            if Y_test[i] == 1:
                predicted_pop.append(X_test[i])
        return predicted_pop

    def calculate_cluster_centroid(self, solution_cluster):
        """
        Calculate the centroid for a given cluster of solutions.
        :param solution_cluster: List of solutions in the cluster
        :return: Cluster centroid
        """
        # Get number of variable shape
        try:
            n_vars = len(solution_cluster[0])
        except TypeError:
            solution_cluster = np.array(solution_cluster)
            return solution_cluster.tolist()

        # TODO: this is lazy garbage fix whats coming in
        cluster = []
        for i in range(len(solution_cluster)):
            # cluster.append(solution_cluster[i].tolist())
            cluster.append(solution_cluster[i])
        solution_cluster = np.asarray(cluster)

        # Get number of solutions
        length = solution_cluster.shape[0]

        centroid_points = []

        # calculate centroid for each variable, by taking mean of every variable of cluster
        for i in range(n_vars):
            # calculate sum over cluster
            centroid_points.append(np.sum(solution_cluster[:, i]))

        return [x / length for x in centroid_points]

    def check_boundaries(self, pop):
        """
        Check and fix the boundaries of the given population.
        :param pop: Population to check and fix boundaries
        :return: Population with corrected boundaries
        """
        # check whether numpy array or pymoo population is given
        if isinstance(pop, Population):
            pop = pop.get("X")

        # check if any solution is outside the bounds
        for individual in pop:
            for i in range(len(individual)):
                if individual[i] > self.problem.xu[i]:
                    individual[i] = self.problem.xu[i]
                elif individual[i] < self.problem.xl[i]:
                    individual[i] = self.problem.xl[i]
        return pop

    def random_strategy(self, N_r):
        """
        Generate a random population within the problem boundaries.
        :param N_r: Number of random solutions to generate
        :return: Randomly generated population
        """
        # generate a random population of size N_r
        # TODO: Check boundaries
        random_pop = self.random_state.random((N_r, self.problem.n_var))

        # check if any solution is outside the bounds
        for individual in random_pop:
            for i in range(len(individual)):
                if individual[i] > self.problem.xu[i]:
                    individual[i] = self.problem.xu[i]
                elif individual[i] < self.problem.xl[i]:
                    individual[i] = self.problem.xl[i]

        return random_pop



import torch
import torch.nn as nn
import torch.optim as optim

from .base import TimeSeriesBase


class PredictorLSTM(nn.Module):
    """Lightweight LSTM network for time-series forecasting.

    Attributes
    ----------
    input_dim : int
        Number of input features per time step.
    hidden_dim : int
        Number of hidden units in LSTM layers.
    num_layers : int
        Number of LSTM layers.
    lstm : nn.LSTM
        Core LSTM network module.
    output_layer : nn.Linear
        Final output projection layer.
    """

    def __init__(self, input_dim, hidden_dim, num_layers, output_dim, dropout):
        """Initialize the LSTM predictor model.

        Parameters
        ----------
        input_dim : int
            Number of input features per time step.
        hidden_dim : int
            Number of hidden units in LSTM layers.
        num_layers : int
            Number of LSTM layers.
        output_dim : int
            Number of output features to predict.
        dropout : float
            Dropout probability between LSTM layers (applied only when num_layers > 1).
        """
        super().__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers

        # LSTM architecture with cuDNN optimization
        self.lstm = nn.LSTM(
            input_size=input_dim,
            hidden_size=hidden_dim,
            num_layers=num_layers,
            batch_first=True,
            dropout=dropout if num_layers > 1 else 0.0,
        )

        # Output projection layer
        self.output_layer = nn.Linear(hidden_dim, output_dim)

    def forward(self, x: torch.Tensor, hidden=None):
        """Forward pass through the LSTM network.

        Parameters
        ----------
        x : torch.Tensor
            Input sequence tensor of shape (batch_size, sequence_length, input_dim).
        hidden : tuple, optional
            Initial hidden state tuple (h_0, c_0) where:
            - h_0 : torch.Tensor, shape (num_layers, batch_size, hidden_dim)
            - c_0 : torch.Tensor, shape (num_layers, batch_size, hidden_dim)

        Returns
        -------
        pred : torch.Tensor
            Prediction for next time step with shape (batch_size, output_dim).
        hidden_out : tuple
            Final hidden state tuple (h_n, c_n) with same shapes as input hidden.
        """
        # LSTM forward pass
        output, hidden_out = self.lstm(x, hidden)

        # Extract hidden state from the last time step
        last_hidden = output[:, -1, :]  # shape: (batch_size, hidden_dim)

        # Generate prediction for next time step
        pred = self.output_layer(last_hidden)  # shape: (batch_size, output_dim)

        return pred, hidden_out


class PredictorMultiLayerLSTM(nn.Module):
    """Multi-layer LSTM model for time series prediction.

    Uses sequential LSTM layers to encode temporal patterns and predict next time step.

    Attributes
    ----------
    input_dim : int
        Number of input features per time step.
    hidden_dim : int
        Number of hidden units in each LSTM layer.
    lstm_layers : nn.ModuleList
        Stack of LSTM layers.
    output_layer : nn.Sequential
        Output network to generate predictions.
    """

    def __init__(self, input_dim, hidden_dim, num_layers, output_dim):
        """Initialize the Multi-layer LSTM predictor model.

        Parameters
        ----------
        input_dim : int
            Number of input features per time step.
        hidden_dim : int
            Number of hidden units in each LSTM layer.
        num_layers : int
            Number of LSTM layers to stack.
        output_dim : int
            Number of output features to predict.
        """
        super().__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim

        # Create stack of LSTM layers
        # First layer: input_dim -> hidden_dim, subsequent layers: hidden_dim -> hidden_dim
        self.lstm_layers = nn.ModuleList(
            [
                nn.LSTM(
                    input_size=input_dim if i == 0 else hidden_dim,
                    hidden_size=hidden_dim,
                    batch_first=True,  # Input shape: (batch_size, sequence_length, input_dim)
                    num_layers=1,  # Each layer is separate for individual hidden state control
                    dropout=0,
                )
                for i in range(num_layers)
            ]
        )

        # Output network to transform LSTM encoding to prediction
        self.output_layer = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim // 2),  # Reduce dimension for computational efficiency
            nn.ReLU(),  # Introduce non-linearity
            nn.Dropout(0.1),  # Prevent overfitting
            nn.Linear(hidden_dim // 2, output_dim),  # Final prediction layer
        )

    def forward(self, x: torch.Tensor, hidden_states=None):
        """Forward pass through the LSTM network.

        Parameters
        ----------
        x : torch.Tensor
            Input sequence tensor of shape (batch_size, sequence_length, input_dim).
        hidden_states : list of tuples, optional
            Previous hidden states for each LSTM layer as (hidden_state, cell_state) tuples.

        Returns
        -------
        prediction : torch.Tensor
            Predicted values for next time step with shape (batch_size, output_dim).
        new_hidden_states : list of tuples
            Updated hidden states for each LSTM layer.
        """
        # Initialize hidden states if not provided
        if hidden_states is None:
            hidden_states = [None] * len(self.lstm_layers)

        # Process input through each LSTM layer sequentially
        output = x  # Start with raw input: shape (batch_size, sequence_length, input_dim)
        new_hidden_states = []  # Store updated hidden states for all layers

        # Pass data through each LSTM layer
        for i, lstm_layer in enumerate(self.lstm_layers):
            # Each LSTM layer processes the sequence
            # output shape: (batch_size, sequence_length, hidden_dim)
            output, (h_n, c_n) = lstm_layer(output, hidden_states[i] if hidden_states[i] is not None else None)
            new_hidden_states.append((h_n, c_n))

        # Extract final time step encoding for prediction
        # last_output shape: (batch_size, hidden_dim)
        last_output = output[:, -1, :]

        # Generate prediction for next time step
        # prediction shape: (batch_size, output_dim)
        prediction = self.output_layer(last_output)

        return prediction, new_hidden_states


class LSTMpredictor(TimeSeriesBase):
    """LSTM-based time series predictor with sliding window approach."""

    def __init__(
        self,
        sequence_length,
        hidden_dim: int = 64,
        num_layers: int = 1,
        dropout: float = 0.0,
        epochs: int = 50,
        batch_size: int = 32,
        lr: float = 0.001,
        device: str = "cpu",
        patience: int = 5,
        seed: int | None = None,
        model_type: str = "lstm",
        incremental_learning: bool = False,
    ):
        """Initialize the LSTM predictor with training configuration.

        Parameters
        ----------
        sequence_length : int
            Number of historical time steps used for prediction.
        hidden_dim : int
            Number of hidden units in LSTM layers, by default 64.
        num_layers : int
            Number of LSTM layers, by default 1.
        dropout : float
            Dropout probability between layers, by default 0.0.
        epochs : int
            Maximum number of training epochs, by default 50.
        batch_size : int
            Number of samples per training batch, by default 32.
        lr : float
            Learning rate for optimizer, by default 0.001.
        device : str
            Computation device ('cpu' or 'cuda'), by default "cpu".
        patience : int
            Early stopping patience (epochs without improvement), by default 5.
        seed : int, optional
            Random seed for reproducibility, by default None.
        model_type : str
            Type of model architecture, by default "lstm".
            Supported values: 'lstm', 'transformer'
        incremental_learning : bool
            Whether to use incremental learning mode, by default False.
            If True, model will be reused and updated with new data.
        """
        self.sequence_length = sequence_length
        if self.sequence_length < 1:
            raise ValueError("The sequence length should be greater than 0.")

        # Model hyperparameters
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        self.dropout = dropout if num_layers > 1 else 0.0

        # Training parameters
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr

        # Device configuration
        self.device = torch.device(device) if device else torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Training parameters
        self.patience = patience
        self._improvement_threshold = 0.001

        # Repeatability
        self.seed = seed
        if self.seed is not None:
            self._set_random_seed(self.seed)

        # Initialize parent class
        super().__init__(self.sequence_length, self.device, model_type=model_type)

        # Incremental learning setup
        self.model = None
        self.incremental_learning = incremental_learning

    def convert_train_predict(self, time_series_data: list[list[float]]) -> torch.Tensor:
        """Complete pipeline: convert data, train model, and generate prediction.

        Parameters
        ----------
        time_series_data : list[list[float]]
            Input time series data as list of feature vectors.
            Shape: (n_timesteps, n_features)
            Example: [[1.0, 2.0], [1.1, 2.1], [1.2, 2.2]] for 3 timesteps, 2 features

        Returns
        -------
        next_prediction : torch.Tensor
            Predicted values for the next time step.
            Shape: (n_features,)
            Example: tensor([1.3, 2.3]) for 2 features

        Notes
        -----
        This method provides an end-to-end workflow:
        1. Convert raw data to tensor format
        2. Train LSTM model on the entire dataset
        3. Generate prediction for the next time step
        The trained model is stored in self.model for future predictions.
        """
        # Convert input data to tensor format
        tensor_data = self.convert_to_tensor(time_series_data)

        # Train model on the prepared data
        trained_model, _ = self.train_model(tensor_data)

        # Generate prediction using trained model
        next_prediction = self.predict_future(trained_model, tensor_data, n_steps=1)

        return next_prediction[0]

    def train_model(self, series_data: torch.Tensor):
        """Train LSTM model on provided time series data.

        Parameters
        ----------
        series_data : torch.Tensor
            Input time series data of shape (n_timesteps, n_features)

        Returns
        -------
        tuple[nn.Module, list[float]]
            Trained model and list of training losses

        Notes
        -----
        Training workflow:
        1. Prepare training sequences and targets using sliding window
        2. Create DataLoader for efficient batch processing
        3. Train model using configured optimization parameters
        """
        # Prepare training sequences and targets
        sequences_tensor, targets_tensor = self.prepare_training_data(series_data)

        # Create DataLoader with appropriate batch size
        dataloader = self.create_training_dataloader(
            sequences_tensor, targets_tensor, batch_size=min(self.batch_size, sequences_tensor.shape[0])
        )

        # Train LSTM model
        trained_model, losses = self._train_model(dataloader)

        return trained_model, losses

    def _construct_model(self, dataloader: torch.utils.data.DataLoader) -> nn.Module:
        """Construct LSTM model using prepared sequential data.

        Parameters
        ----------
        dataloader : torch.utils.data.DataLoader
            DataLoader containing training sequences and corresponding targets.

        Returns
        -------
        model : nn.Module
            Configured LSTM model instance.
        """
        # Extract model dimensions from dataset
        input_dim = dataloader.dataset[0][0].shape[-1]
        output_dim = dataloader.dataset[0][1].shape[-1]

        # Initialize model based on continual learning configuration
        if (self.model is None) or (not self.incremental_learning):
            # Create new model for initial training or non-continual learning scenarios
            model = PredictorLSTM(
                input_dim=input_dim,
                hidden_dim=self.hidden_dim,
                num_layers=self.num_layers,
                output_dim=output_dim,
                dropout=self.dropout,
            ).to(self.device)
            if self.incremental_learning:
                self.model = model  # Store model reference for future continual learning
        else:
            # Utilize existing model for continual learning
            model = self.model

        return model

    def _train_model(self, dataloader: torch.utils.data.DataLoader) -> tuple[nn.Module, list[float]]:
        """Train LSTM model on prepared sequence data.

        Parameters
        ----------
        dataloader : torch.utils.data.DataLoader
            DataLoader containing training sequences and targets

        Returns
        -------
        model : nn.Module
            Trained LSTM model instance
        losses : list[float]
            List of average losses per epoch

        Notes
        -----
        Training configuration:
        - Loss: Mean Squared Error (MSE)
        - Optimizer: Adam with L2 regularization
        - Learning rate scheduler: StepLR with decay
        - Gradient clipping: Prevents gradient explosion
        - Early stopping: Monitors validation loss improvement

        The training process includes:
        1. Model initialization with appropriate dimensions
        2. Batch-wise forward and backward passes
        3. Learning rate scheduling
        4. Early stopping based on loss improvement
        5. Progress logging at regular intervals
        """
        # Construct model
        model = self._construct_model(dataloader)

        # Training configuration
        criterion, optimizer, scheduler = self._optimizer_scheduler(model)

        # Early stopping initialization
        best_loss = float("inf")  # positive infinity
        patience_counter = 0

        # Training loop
        losses = []
        for epoch in range(self.epochs):
            model.train()
            epoch_loss = 0.0
            batch_count = 0

            for batch_idx, (X_batch, y_batch) in enumerate(dataloader):
                X_batch, y_batch = X_batch.to(self.device), y_batch.to(self.device)

                # Forward pass
                optimizer.zero_grad()
                predictions, _ = model(X_batch)  # stateless; independent-sequence
                loss = criterion(predictions, y_batch)

                # Backward pass with gradient clipping
                loss.backward()
                torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0)
                optimizer.step()

                # Accumulate loss
                current_loss = loss.item()
                epoch_loss += current_loss
                batch_count += 1

                # Print batch progress
                if (batch_idx + 1) % 100 == 0:
                    print(f"Batch {batch_idx}, Loss: {current_loss:.4f}")

            # Update learning rate
            scheduler.step()

            # Calculate average epoch loss
            avg_epoch_loss = epoch_loss / batch_count if batch_count > 0 else 0.0
            losses.append(avg_epoch_loss)

            # Early stopping check
            if avg_epoch_loss < best_loss * (1 - self._improvement_threshold):  # relative improvement
                best_loss = avg_epoch_loss
                patience_counter = 0
            else:
                patience_counter += 1
                if patience_counter >= self.patience:
                    # print(f"Early stopping at epoch {epoch + 1}")
                    break

            # Print epoch progress
            if (epoch + 1) % 100 == 0:
                current_lr = optimizer.param_groups[0]["lr"]
                print(f"Epoch [{epoch + 1}/{self.epochs}], Average Loss: {avg_epoch_loss:.6f}, LR: {current_lr:.6f}")

        # Store model for future continual learning
        if self.incremental_learning:
            self.model = model
        return model, losses

    def _optimizer_scheduler(self, model: nn.Module):
        """Configure optimizer and learning rate scheduler.

        Parameters
        ----------
        model : torch.nn.Module
            The model to optimize

        Returns
        -------
        tuple
            criterion: loss function
            optimizer: configured optimizer
            scheduler: learning rate scheduler
        """
        criterion = nn.MSELoss()
        optimizer = optim.Adam(model.parameters(), lr=self.lr, weight_decay=1e-5)
        scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=10, gamma=0.9)
        return criterion, optimizer, scheduler



import numpy as np
from pymoo.core.population import Population
from pymoo.operators.survival.rank_and_crowding import RankAndCrowding

from pydmoo.algorithms.base.dmoo.dmoead import DMOEAD
from pydmoo.core.sample_gaussian import univariate_gaussian_sample


class MOEADKTMM(DMOEAD):
    """Knowledge Transfer with Mixture Model.

    Zou, J., Hou, Z., Jiang, S., Yang, S., Ruan, G., Xia, Y., and Liu, Y. (2025).
    Knowledge transfer with mixture model in dynamic multi-objective optimization.
    IEEE Transactions on Evolutionary Computation, in press.
    https://doi.org/10.1109/TEVC.2025.3566481
    """

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.size_pool = 14  # the size of knowledge pool
        self.denominator = 0.5

    def _response_mechanism(self) -> Population:
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # sample self.pop_size solutions in decision space
        samples_old = self.sampling_new_pop()

        # select self.pop_size/2 individuals with better convergence and diversity
        samples = samples_old[:int(len(samples_old)/2)]

        # knowledge in decision space
        means_stds_ps, mean, std = self._in_decision_or_objective_space_1d(samples, "decision_space")
        mean_new, std_new = self._select_means_stds(means_stds_ps, mean, std)

        # sample self.pop_size solutions in decision space
        X = univariate_gaussian_sample(mean_new, std_new, self.pop_size, random_state=self.random_state)

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X = np.clip(X, xl, xu)  # not provided in the original reference literature

        # merge
        pop = Population.merge(samples_old, Population.new(X=X))

        return pop

    def _in_decision_or_objective_space_1d(self, samples, decision_or_objective="decision_space"):
        # decision space or objective space
        flag = "X" if decision_or_objective == "decision_space" else "F"

        means_stds = self.data.get("means_stds", [])

        flag_value = self.opt.get(flag)
        if len(flag_value) <= 1:
            flag_value = self.pop.get(flag)
            flag_value = flag_value[:2]

        means_stds.append((np.mean(flag_value, axis=0), np.std(flag_value, axis=0), self.n_iter - 1))  # 1-based
        self.data["means_stds"] = means_stds

        flag_value = samples.get(flag)
        mean, std = np.mean(flag_value, axis=0), np.std(flag_value, axis=0)
        return means_stds, mean, std

    def sampling_new_pop(self):
        samples = self.initialization.sampling(self.problem, self.pop_size)
        samples = self.evaluator.eval(self.problem, samples)

        # do a survival to recreate rank and crowding of all individuals
        samples = RankAndCrowding().do(self.problem, samples, n_survive=len(samples))
        return samples

    def _select_means_stds(self, means_stds, mean_new, std_new):
        # Unpack means and stds
        means = np.array([m[0] for m in means_stds])
        stds = np.array([m[1] for m in means_stds])

        # Calculate distances
        mean_diffs = means - mean_new
        std_diffs = stds - std_new

        distances = np.sqrt(np.sum(mean_diffs**2, axis=1) + np.sum(std_diffs**2, axis=1))

        # Get top K closest
        top_k_idx = np.argsort(distances)[:self.size_pool]
        top_k_dist = distances[top_k_idx]
        top_k_means = means[top_k_idx]
        top_k_stds = stds[top_k_idx]

        # Update pool
        self._update_means_stds_pool(means_stds, top_k_idx)

        # Calculate weights
        weights = 1 / (top_k_dist + 1e-8)  # Add small epsilon to avoid division by zero
        weights = weights / (np.sum(weights) + self.denominator)

        # Weighted combination
        mean_new = (1 - np.sum(weights)) * mean_new + np.sum(weights[:, None] * top_k_means, axis=0)
        std_new = (1 - np.sum(weights)) * std_new + np.sum(weights[:, None] * top_k_stds, axis=0)
        return mean_new, std_new

    def _update_means_stds_pool(self, means_stds, top_k_idx) -> None:
        self.data["means_stds"] = [means_stds[i] for i in top_k_idx]
        return None



import numpy as np


def igp_based_predictor(PS, PF, F_, sigma_n_2):
    # algorithm 2: IGP-Based Predictor
    # algorithm 1: IGPR
    K_g = np.dot(PF, PF.T)
    K_g_noise = K_g + sigma_n_2 * np.eye(len(PF))
    k_inv = np.linalg.inv(K_g_noise)
    # k_inv = cholesky_inverse_numpy(K_g_noise)

    X_ = np.dot(np.dot(np.dot(F_, PF.T), k_inv), PS)
    return X_


def cholesky_inverse_numpy(A):
    """
    Compute the inverse of a symmetric positive definite (SPD) matrix using Cholesky decomposition.

    Parameters
    ----------
    A : numpy.ndarray
        A symmetric positive definite matrix (must satisfy A = A.T and all eigenvalues > 0)

    Returns
    -------
    numpy.ndarray
        The inverse matrix A^{-1} computed via Cholesky decomposition

    Raises
    ------
    AssertionError
        If input matrix is not symmetric or not positive definite
    """
    # 1. Check if matrix is symmetric and positive definite
    assert np.allclose(A, A.T), "Matrix must be symmetric"
    assert np.all(np.linalg.eigvals(A) > 0), "Matrix must be positive definite"

    # 2. Compute Cholesky decomposition A = L L^T
    # L is lower triangular matrix with positive diagonal entries
    L = np.linalg.cholesky(A)

    # 3. Compute inverse of L (triangular matrix inversion)
    # Since L is lower triangular, its inverse can be computed efficiently
    inv_L = np.linalg.inv(L)  # Alternative: solve triangular systems

    # 4. Compute inverse of A using A^{-1} = (L^{-1})^T L^{-1}
    A_inv = inv_L.T @ inv_L

    return A_inv



import numpy as np


def do_degeneration(PF, F, eps=1e-10):
    min_pf = np.min(PF, axis=0)
    max_pf = np.max(PF, axis=0)

    mask = (max_pf - min_pf) >= eps
    PF = PF[:, mask]
    F = F[:, mask]
    return PF, F


def failure_count(PF, F):
    min_pf = np.min(PF, axis=0)
    max_pf = np.max(PF, axis=0)

    # Check if any element in each row exceeds the bounds
    # (data < lb) creates boolean matrix for lower bound violations
    # (data > ub) creates boolean matrix for upper bound violations
    # | combines both violations (OR operation)
    # np.any(..., axis=1) checks if any violation exists in each row
    out_of_bounds_mask = np.any((F < min_pf) | (F > max_pf), axis=1)

    # Count number of invalid solutions (rows with at least one violation)
    # True values are counted as 1, False as 0
    num_invalid = np.sum(out_of_bounds_mask)

    # Get indices of invalid rows
    # np.where returns a tuple (we take first element with [0])
    invalid_indices = np.where(out_of_bounds_mask)[0]

    valid_indices = np.where(~out_of_bounds_mask)[0]
    return num_invalid, valid_indices, invalid_indices


def matrix_conditional_update(x_curr, lb, ub, x_prev):
    """
    Vectorized conditional matrix update with bounded interpolation.

    Parameters
    ----------
    x_curr : ndarray, shape (N, n)
        Observation matrix.
    lb : ndarray, shape (1, n) or (n,)
        Lower bounds for each dimension.
    ub : ndarray, shape (1, n) or (n,)
        Upper bounds for each dimension.
    x_prev : ndarray, shape (N, n)
        Previous state matrix.

    Returns
    -------
    ndarray, shape (N, n)
        Updated matrix where:
        - Values within bounds remain unchanged
        - Values below bounds become 0.5*(a + x_prev)
        - Values above bounds become 0.5*(b + x_prev)

    Notes
    -----
    Uses NumPy broadcasting for efficient vectorized operations.
    """
    lb = np.reshape(lb, (1, -1))
    ub = np.reshape(ub, (1, -1))
    x_new = np.zeros_like(x_curr)

    mask = (x_curr >= lb) & (x_curr <= ub)
    x_new[mask] = x_curr[mask]

    mask = x_curr < lb
    x_new[mask] = 0.5 * (lb + x_prev)[mask]

    mask = x_curr > ub
    x_new[mask] = 0.5 * (ub + x_prev)[mask]

    return x_new


def clip_and_randomize(x, lb, ub, random_state=None):
    """
    Clip values to bounds with random replacement for out-of-bounds values.

    Parameters
    ----------
    x : ndarray, shape (N, n)
        Input matrix.
    lb : ndarray, shape (1, n) or (n,)
        Lower bounds.
    ub : ndarray, shape (1, n) or (n,)
        Upper bounds.

    Returns
    -------
    ndarray, shape (N, n)
        Matrix where out-of-bounds values are replaced with uniform random
        values within [lb, ub] for each dimension.

    See Also
    --------
    numpy.random.uniform : Used for random value generation.
    """
    out_of_bounds = (x < lb) | (x > ub)
    random_samples = random_state.uniform(low=lb, high=ub, size=x.shape)
    return np.where(out_of_bounds, random_samples, x)


def clip_by_numpy(x, lb, ub):
    """
    Clip values to interval [lb, ub].

    Parameters
    ----------
    x : ndarray
        Array containing elements to clip.
    lb : ndarray or scalar
        Minimum value.
    ub : ndarray or scalar
        Maximum value.

    Returns
    -------
    ndarray
        Clipped array where:
        x_clipped = max(lb, min(ub, x))

    Notes
    -----
    This is a wrapper around numpy.clip() with identical behavior.
    """
    return np.clip(x, lb, ub)



import numpy as np


def reconstruct_covariance_from_triu(triu_elements, n_features):
    """
    Reconstruct a symmetric covariance matrix from upper triangular elements.

    This function takes the flattened upper triangular elements (including diagonal)
    of a covariance matrix and reconstructs the full symmetric matrix by leveraging
    the symmetry property of covariance matrices.

    Parameters
    ----------
    triu_elements : array_like
        Flattened array containing the upper triangular elements of the covariance matrix,
        including the diagonal elements. The elements should be in row-major order.
        Shape: (n_triu_elements,) where n_triu_elements = n_features * (n_features + 1) // 2
    n_features : int
        Dimensionality of the feature space, which determines the size of the
        output covariance matrix (n_features x n_features)

    Returns
    -------
    cov_matrix : ndarray
        Reconstructed symmetric covariance matrix of shape (n_features, n_features).
        The matrix satisfies cov_matrix[i, j] == cov_matrix[j, i] for all i, j.

    Raises
    ------
    ValueError
        If the length of triu_elements does not match the expected number of
        upper triangular elements for the given n_features

    Notes
    -----
    The number of upper triangular elements (including diagonal) for an n x n matrix
    is given by: n_triu_elements = n * (n + 1) // 2

    Examples
    --------
    >>> triu_elements = np.array([1.0, 0.5, 0.3, 2.0, 0.4, 3.0])
    >>> n_features = 3
    >>> reconstruct_covariance_from_triu(triu_elements, n_features)
    array([[1. , 0.5, 0.3],
           [0.5, 2. , 0.4],
           [0.3, 0.4, 3. ]])
    """
    # Validate input dimensions
    expected_triu_elements = n_features * (n_features + 1) // 2
    if len(triu_elements) != expected_triu_elements:
        raise ValueError(
            f"Invalid number of triu_elements: expected {expected_triu_elements}, "
            f"got {len(triu_elements)}"
        )

    # Initialize output matrix with zeros
    # Shape: (n_features, n_features)
    cov_matrix = np.zeros((n_features, n_features), dtype=np.float64)

    # Get indices for upper triangular portion (including diagonal)
    # rows, cols: arrays of indices where row <= col
    rows, cols = np.triu_indices(n_features)

    # Fill upper triangular portion with input elements
    # This sets cov_matrix[i, j] for all i <= j
    cov_matrix[rows, cols] = triu_elements

    # Fill lower triangular portion using symmetry
    # Since covariance matrix is symmetric: cov_matrix[i, j] = cov_matrix[j, i]
    # This sets cov_matrix[j, i] = cov_matrix[i, j] for all i < j
    mask = rows != cols
    cov_matrix[cols[mask], rows[mask]] = cov_matrix[rows[mask], cols[mask]]

    return cov_matrix


def make_semidefinite(matrix, tol=1e-8):
    """
    Convert a symmetric matrix to positive semi-definite form.

    This function takes a symmetric matrix and ensures it is positive semi-definite
    by eigen-decomposition and thresholding of negative eigenvalues. The input matrix
    is first symmetrized, then any negative eigenvalues are set to zero.

    Parameters
    ----------
    matrix : array_like
        Input matrix to be converted to positive semi-definite form.
        Should be square and approximately symmetric.
        Shape: (n, n)
    tol : float, optional
        Tolerance for eigenvalue thresholding and symmetry check.
        Eigenvalues less than this value will be set to zero.
        Default: 1e-8

    Returns
    -------
    psd_matrix : ndarray
        Positive semi-definite matrix with the same shape as input.
        All eigenvalues are guaranteed to be >= 0.
        Shape: (n, n)

    Raises
    ------
    LinAlgError
        If the eigen-decomposition fails (e.g., matrix is not numerically symmetric)
    ValueError
        If input matrix is not square

    Notes
    -----
    The algorithm proceeds in three steps:
    1. Symmetrize the input matrix: (matrix + matrix.T) / 2
    2. Compute eigen-decomposition using eigh (optimized for symmetric matrices)
    3. Threshold negative eigenvalues to zero and reconstruct

    This method preserves the eigenvectors while modifying only the eigenvalues
    that violate positive semi-definiteness.

    Examples
    --------
    >>> import numpy as np
    >>> matrix = np.array([[1.0, 0.5], [0.5, 1.0]])
    >>> make_semidefinite(matrix)
    array([[1. , 0.5],
           [0.5, 1. ]])

    >>> # Example with negative eigenvalues
    >>> matrix_indefinite = np.array([[1.0, 2.0], [2.0, 1.0]])
    >>> print("Original eigenvalues:", np.linalg.eigvalsh(matrix_indefinite))
    Original eigenvalues: [-1.  3.]
    >>> psd_matrix = make_semidefinite(matrix_indefinite)
    >>> print("PSD eigenvalues:", np.linalg.eigvalsh(psd_matrix))
    PSD eigenvalues: [0. 3.]

    See Also
    --------
    numpy.linalg.eigh : Eigen decomposition for symmetric/Hermitian matrices
    numpy.diag : Create diagonal matrix from vector
    """
    # Input validation
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError(f"Input matrix must be square, got shape {matrix.shape}")

    # Step 1: Ensure matrix is symmetric
    # Average with transpose to eliminate numerical asymmetry
    # Result: symmetric_matrix = (matrix + matrix.T) / 2
    symmetric_matrix = (matrix + matrix.T) / 2

    # Step 2: Compute eigen-decomposition
    # eigh is preferred over eig for symmetric matrices (faster, more stable)
    # Returns:
    #   eigvals : 1D array of eigenvalues in ascending order
    #   eigvecs : 2D array where columns are corresponding eigenvectors
    eigvals, eigvecs = np.linalg.eigh(symmetric_matrix)

    # Step 3: Threshold negative eigenvalues to zero
    # Preserve eigenvalues >= tol, set others to 0
    # This ensures positive semi-definiteness while minimizing distortion
    eigvals_psd = np.where(eigvals < tol, 0.0, eigvals)

    # Step 4: Reconstruct matrix from modified eigenvalues
    # Matrix = V * diag(λ) * V.T where V contains eigenvectors as columns
    psd_matrix = eigvecs @ np.diag(eigvals_psd) @ eigvecs.T

    return psd_matrix



import numpy as np
from pymoo.core.population import Population
from pymoo.operators.survival.rank_and_crowding import RankAndCrowding

from pydmoo.algorithms.modern.moeadde_imkt_n import MOEADDEIMKTN
from pydmoo.algorithms.modern.nsga2_imkt_n_lstm import prepare_data_mean_cov
from pydmoo.algorithms.utils.utils import make_semidefinite, reconstruct_covariance_from_triu
from pydmoo.core.bounds import clip_and_randomize
from pydmoo.core.inverse import closed_form_solution
from pydmoo.core.lstm.lstm import LSTMpredictor
from pydmoo.core.sample_gaussian import multivariate_gaussian_sample


class MOEADDEIMNcLSTM(MOEADDEIMKTN):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._n_timesteps = 10
        self._sequence_length = 5  # Use 5 historical time steps to predict next step
        self._incremental_learning = False

    def _setup(self, problem, **kwargs):
        super()._setup(problem, **kwargs)

        # Must be here
        self._lstm = LSTMpredictor(
            self._sequence_length,
            hidden_dim=64,
            num_layers=1,
            epochs=50,
            batch_size=32,
            lr=0.001,
            device="cpu",  # for fair comparison
            patience=5,
            seed=self.seed,
            model_type="lstm",
            incremental_learning=self._incremental_learning,
        )

    def _response_mechanism(self) -> Population:
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # sample self.pop_size individuals in decision space
        samples_old = self.sampling_new_pop()

        # select self.pop_size/2 individuals with better convergence and diversity
        samples = samples_old[:int(len(samples_old)/2)]

        # knowledge in objective space
        means_covs, mean, cov = self._in_decision_or_objective_space_nd(samples, "objective_space")

        # Check if sufficient historical data is available for LSTM prediction
        if len(means_covs) > self._n_timesteps:
            # Update pool
            self.data["means_covs"] = means_covs[self._n_timesteps:]

            # Prepare time series data from historical means and covariance matrices
            time_series_data = prepare_data_mean_cov(self._n_timesteps, means_covs)

            # Initialize predictor and generate prediction for next time step
            next_prediction = self._lstm.convert_train_predict(time_series_data)

            # Convert prediction tensor to numpy array for further processing
            next_prediction = next_prediction.numpy()

            # Split prediction into mean and covariance components
            # First n_obj elements represent the mean vector, Remaining elements represent the flattened covariance matrix values
            mean_new, cov_new_ = next_prediction[:self.problem.n_obj], next_prediction[self.problem.n_obj:]
            cov_new = reconstruct_covariance_from_triu(cov_new_, len(mean_new))
            cov_new = make_semidefinite(cov_new)

        else:
            mean_new, cov_new = self._select_means_covs(means_covs, mean, cov)

        # sample self.pop_size individuals in objective space
        F = multivariate_gaussian_sample(mean_new, cov_new, self.pop_size, random_state=self.random_state)

        # TODO
        # inverse mapping
        # X = FB
        B = closed_form_solution(samples.get("X"), samples.get("F"))

        # X = FB
        X = np.dot(F, B)

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X = clip_and_randomize(X, xl, xu, random_state=self.random_state)

        # merge
        pop = Population.merge(samples_old, Population.new(X=X))

        return pop

    def sampling_new_pop(self):
        ps = self.opt.get("X")
        X = self.pop.get("X")

        if not self.problem.has_constraints():

            last_ps = self.data.get("last_ps", [])
            if len(last_ps) == 0:
                last_ps = ps
            self.data["last_ps"] = ps

            d = np.mean(ps, axis=0) - np.mean(last_ps, axis=0)

            radius = max(np.linalg.norm(d) / self.problem.n_obj, 0.1)

            X = X + d + self.random_state.uniform(low=-radius, high=radius, size=X.shape)

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X = clip_and_randomize(X, xl, xu, random_state=self.random_state)

        samples = Population.new(X=X)
        samples = self.evaluator.eval(self.problem, samples)

        # do a survival to recreate rank and crowding of all individuals
        samples = RankAndCrowding().do(self.problem, samples, n_survive=len(samples))
        return samples

    def _select_means_covs(self, means_covs, mean_new, cov_new):
        # Unpack means and stds
        means = np.array([m[0] for m in means_covs])
        covs = np.array([m[1] for m in means_covs])

        # Weighted combination
        mean_new = 0.5 * mean_new + 0.5 * means[-1]
        cov_new = 0.5 * cov_new + 0.5 * covs[-1]

        return mean_new, cov_new


class MOEADDEIMNicLSTM(MOEADDEIMNcLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._n_timesteps = 10
        self._sequence_length = 5  # Use 5 historical time steps to predict next step
        self._incremental_learning = True



import numpy as np
from pymoo.core.population import Population

from pydmoo.algorithms.modern.nsga2_imkt import NSGA2IMKT
from pydmoo.core.bounds import clip_and_randomize
from pydmoo.core.inverse import closed_form_solution
from pydmoo.core.lstm.lstm import LSTMpredictor
from pydmoo.core.sample_gaussian import univariate_gaussian_sample


class NSGA2IMLSTM(NSGA2IMKT):
    """Inverse Modeling with LSTM (IMLSTM).

    Inverse Modeling for Dynamic Multiobjective Optimization with LSTM prediction In objective Space.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5

        self._n_timesteps = 10
        self._sequence_length = 5  # Use 5 historical time steps to predict next step
        self._incremental_learning = False

    def _setup(self, problem, **kwargs):
        super()._setup(problem, **kwargs)

        # Must be here
        self._lstm = LSTMpredictor(
            self._sequence_length,
            hidden_dim=64,
            num_layers=1,
            epochs=50,
            batch_size=32,
            lr=0.001,
            device="cpu",  # for fair comparison
            patience=5,
            seed=self.seed,
            model_type="lstm",
            incremental_learning=self._incremental_learning,
        )

    def _response_mechanism(self) -> Population:
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # sample self.pop_size individuals in decision space
        samples_old = self.sampling_new_pop()

        # select self.pop_size/2 individuals with better convergence and diversity
        samples = samples_old[:int(len(samples_old)/2)]

        # knowledge in objective space
        means_stds, mean, std = self._in_decision_or_objective_space_1d(samples, "objective_space")

        # Check if sufficient historical data is available for LSTM prediction
        if len(means_stds) > self._n_timesteps:
            # Update pool
            self.data["means_stds"] = means_stds[self._n_timesteps:]

            # Prepare time series data from historical means and standard deviations
            time_series_data = prepare_data_means_std(self._n_timesteps, means_stds)

            # Initialize predictor and generate prediction for next time step
            next_prediction = self._lstm.convert_train_predict(time_series_data)

            # Convert prediction tensor to numpy array for further processing
            next_prediction = next_prediction.numpy()

            # Split prediction into mean and standard deviation components
            # First n_obj elements represent mean values, remaining elements represent standard deviations
            mean_new, std_new = next_prediction[:self.problem.n_obj], next_prediction[self.problem.n_obj:]
            std_new = np.abs(std_new)

        else:
            mean_new, std_new = self._select_means_stds(means_stds, mean, std)

        # sample self.pop_size individuals in objective space
        F = univariate_gaussian_sample(mean_new, std_new, self.pop_size, random_state=self.random_state)

        # TODO
        # inverse mapping
        # X = FB
        B = closed_form_solution(samples.get("X"), samples.get("F"))

        # X = FB
        X = np.dot(F, B)

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X = clip_and_randomize(X, xl, xu, random_state=self.random_state)

        # merge
        pop = Population.merge(samples_old, Population.new(X=X))

        return pop


def prepare_data_means_std(n_timesteps, means_stds):
    """Prepare time series data from means and standard deviations.

    This function converts a sequence of mean vectors and standard deviation vectors
    into a time series format suitable for machine learning models. It concatenates
    mean values and standard deviation values to create feature vectors for each time step.

    Parameters
    ----------
    means_stds : list of tuples
        List containing (mean, std, n_iter) pairs for each time step, where:
        - mean: 1D numpy array of mean values
        - std: 1D numpy array of standard deviation values
        - n_iter: number of iterations (not used in feature extraction)

    Returns
    -------
    time_series_data : list
        Combined feature data with shape (n_timesteps, n_features)
        Each row represents a time step containing:
        [mean_1, mean_2, ..., mean_n, std_1, std_2, ..., std_n]
    """
    # Create time series data
    time_series_data = []  # shape: (n_timesteps, n_features)

    # Process only the most recent n_timesteps
    for m, s, _ in means_stds[-n_timesteps:]:
        # Combine mean vector and standard deviation vector
        # [*m] unpacks all mean values
        # [*s] unpacks all standard deviation values
        feature_vector = [*m, *s]

        time_series_data.append(feature_vector)

    return time_series_data


class NSGA2IMiLSTM(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5
        self._n_timesteps = 10
        self._sequence_length = 5
        self._incremental_learning = True


class NSGA2IMiLSTM1003(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5
        self._n_timesteps = 10
        self._sequence_length = 3
        self._incremental_learning = True


class NSGA2IMiLSTM1005(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5
        self._n_timesteps = 10
        self._sequence_length = 5
        self._incremental_learning = True


class NSGA2IMiLSTM1007(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5
        self._n_timesteps = 10
        self._sequence_length = 7
        self._incremental_learning = True


class NSGA2IMiLSTM1009(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5
        self._n_timesteps = 10
        self._sequence_length = 9
        self._incremental_learning = True


class NSGA2IMiLSTM1503(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 15
        self.denominator = 0.5
        self._n_timesteps = 15
        self._sequence_length = 3
        self._incremental_learning = True


class NSGA2IMiLSTM1505(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 15
        self.denominator = 0.5
        self._n_timesteps = 15
        self._sequence_length = 5
        self._incremental_learning = True


class NSGA2IMiLSTM1507(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 15
        self.denominator = 0.5
        self._n_timesteps = 15
        self._sequence_length = 7
        self._incremental_learning = True


class NSGA2IMiLSTM1509(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 15
        self.denominator = 0.5
        self._n_timesteps = 15
        self._sequence_length = 9
        self._incremental_learning = True


class NSGA2IMiLSTM1511(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 15
        self.denominator = 0.5
        self._n_timesteps = 15
        self._sequence_length = 11
        self._incremental_learning = True


class NSGA2IMiLSTM1513(NSGA2IMLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 15
        self.denominator = 0.5
        self._n_timesteps = 15
        self._sequence_length = 13
        self._incremental_learning = True



import numpy as np
from pymoo.core.population import Population
from sklearn.svm import SVR

from pydmoo.algorithms.base.dmoo.dmoead import DMOEAD


class MOEADSVR(DMOEAD):
    """Support Vector Regression (SVR).

    Notes
    -----
    [Official Python Code](https://github.com/LeileiCao/MOEA-D-SVR/blob/master/MOEAD-SVR%20.py)

    References
    ----------
    Cao, L., Xu, L., Goodman, E. D., Bao, C., and Zhu, S. (2020).
    Evolutionary dynamic multiobjective optimization assisted by a support vector regression predictor.
    IEEE Transactions on Evolutionary Computation, 24(2), 305–319.
    https://doi.org/10.1109/TEVC.2019.2925722
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # SVR
        self._q = 4  # the number of preceding values that are correlated with the target value (dimension of input samples in the SVR model)
        self._C = 1000  # the regularization constant in SVR model
        self._epsilon = 0.05  # the insensitive tube size in SVR model
        # self._gamma = 1/d  # the Gaussian RBF kernel parameter used in SVR model, and d is the number of variables

    def _response_mechanism(self) -> Population:
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        old = self.data.get("stacked_X", None)
        if old is None:
            stacked_X = np.expand_dims(X, axis=0)
        else:
            stacked_X = np.concatenate((old, np.expand_dims(X, axis=0)), axis=0)
        self.data["stacked_X"] = stacked_X

        N, d = X.shape
        sol = np.zeros((N, d))
        K = len(stacked_X)

        if K < self._q + 2:
            # recreate the current population without being evaluated
            # Re-evaluate the current population, and update the reference point
            pop = Population.new(X=X)

            return pop

        # Precompute sliding window indices to avoid redundant calculations
        window_indices = np.lib.stride_tricks.sliding_window_view(np.arange(K), self._q + 1)

        for i in range(N):
            for j in range(d):
                # Extract the time series for this (i,j) position
                ts = stacked_X[:K, i, j]

                # Create training data using vectorized sliding windows
                train = ts[window_indices]
                x_train = train[:, :-1]
                y_train = train[:, -1]

                # Train SVR model (consider moving this outside loops if possible)
                # gamma if 'auto', uses 1 / n_features (not provided in code but provided in paper)
                # versionchanged:: 0.22
                # The default value of ``gamma`` changed from 'auto' to 'scale'.
                svr = SVR(kernel='rbf', epsilon=self._epsilon, C=self._C, gamma=1/d)
                model = svr.fit(x_train, y_train)

                # Make prediction
                sol[i, j] = model.predict(ts[-self._q:].reshape(1, -1))

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            sol = np.clip(sol, xl, xu)  # provided in the original reference literature

        # recreate the current population without being evaluated
        pop = Population.new(X=sol)

        return pop
